/*
 * failover -- The failover cmd tool.
 *
 * Copyright (C) 2009, OpenVox
 *
 * mark.liu <mark.liu@openvox.cn>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "failover.h"

static int failover_read(struct failover_dev *dev, char *buf);
static void *failover_kick(void *arg);
static void failover_event_run(struct failover_dev *dev, char *event);

int failover_init(struct failover_dev *dev) {
	if (!dev)
	{
		return -1;
	}

	/* initialize dev */
	memset(dev, 0, sizeof(failover_dev_t));
					
	/* set failover tty_name */
	strncpy(dev->tty.tty_name, DEFAULT_TTY_DEVICE, MAX_DEVICE_NAME - 1);
	
	/* set kick time */
	dev->kick_time_interval = DEFAULT_INTERVAL;
	
	dev->last_event = FAILOVER_EVENT_STOPPED;
	dev->running = 0;

	strncpy(dev->autorun, ENABLE_AUTORUN, MAX_PARAM_NAME - 1);
	
	return 0;
}

int failover_config_tty(struct failover_dev *dev) {
	tty_struct_t *tty = &(dev->tty);
	char *speed = DEFAULT_SPEED;
	char *parms = DEFAULT_PARMS;
	int flow = 0;
	int modem = 0;
	int qflag = 0;
	struct termios serialti; 
	struct termios tempti;
	int ec = 0;
	int fd;
	speed_t baudrate;
	
	if (!dev) {
		fprintf(stderr, "Could not find tty device!\n");
		return -1;
	}
	dev->tty.fd = -1;
	
	/* open tty device */
	/*fd = open(tty->tty_name, O_RDWR|O_NONBLOCK|O_NOCTTY);*/
	fd = open(tty->tty_name, O_RDWR|O_NOCTTY);
	if (fd < 0) {
		fprintf(stderr, "open %s error!\n", tty->tty_name);
		failover_event_run(dev, FAILOVER_EVENT_ERROR);
		return -1;
	}
	
	/* get tty attr */
	if (tcgetattr(fd, &serialti)) {
		ec = EX_OSERR;
		fprintf(stderr, "tcgetattr(%s) error!\n", tty->tty_name);
		goto error;
	}
	
	/* initialize tempti */
	memcpy(&tempti, &serialti, sizeof(tempti));
	
	cfmakeraw(&tempti);
	tempti.c_cc[VMIN] = 1;
	tempti.c_cc[VTIME] = 0;
	
	/* get speed */
	baudrate = parse_baurd(speed);
	if (cfsetispeed(&tempti, baudrate) && cfsetospeed(&tempti, baudrate)) {
		ec = EX_OSERR;
		fprintf(stderr, "cfsetspeed(%s) error!\n", tty->tty_name);
		goto error;
	}
	dev->tty.baudrate = baudrate;
	
	/* get parms, flow, modem */
	if (parse_parms(&tempti.c_cflag, parms, flow, modem)) {
		ec = EX_USAGE;
		goto error;
	}
	
	/* set tty device */
	if (tcsetattr(fd, TCSANOW, &tempti)) {
		ec = EX_OSERR;
		fprintf(stderr, "cftcsetattr(%s)\n", tty->tty_name);
		goto error;
	}
	
	/* print tty device configure info */
	if (!qflag) {
		if (tcgetattr(fd, &tempti)) {
			close(fd);
			fprintf(stderr, "tcgetattr(%s)\n", tty->tty_name);
		}
		print_parms(&tempti, tty->tty_name);
		fflush(stderr);
	}
	cfmakeraw(&tempti);
	
	dev->tty.fd = fd;
	return fd;
error:
	if (fd >= 0) {
		tcsetattr(fd, TCSAFLUSH, &serialti);
		close(fd);
	}
	if (!qflag) {
		fprintf(stderr, "Connection closed.\n");
	}
	failover_event_run(dev, FAILOVER_EVENT_ERROR);
	
	return -1;
}

int failover_free_tty(struct failover_dev *dev) {
	if (!dev) {
		return 0;
	}
	
	int fd = dev->tty.fd;
	if ( fd >= 0) {
		close(fd);
		dev->tty.fd = -1;
	}
	
	return 0;
}

static void failover_event_run(struct failover_dev *dev, char *event) {
	if (!dev)
	{
		return;
	}

	dev->last_event = event;
	
	if (failover_strlen_zero(dev->event_cmd)) {
		return;
	}

	int pid = fork();
	
	switch(pid) {
	case -1: 
		fprintf(stderr, "Could not fork event_cmd for failover '%s': %s\n", dev->dev_name, strerror(errno));
		break;
	case 0:
		execlp(dev->event_cmd, dev->event_cmd, dev->dev_name, event, NULL);
		fprintf(stderr, "Could not run event_cmd for failover '%s': %s\n", dev->dev_name, strerror(errno));
		
		exit(-1);
	}
}

static int failover_read(struct failover_dev *dev, char *buf) {
	int i = 0;
	int num;
	char tmpbuf[2];
	
	if (!dev)
	{
		return -1;
	}

	if (dev->tty.fd) {
		while((num = read(dev->tty.fd, tmpbuf, 1)) > 0) {
			if (tmpbuf[0] == '\n') {
				tmpbuf[0] = tmpbuf[1] = '\0';
				break;
			} else {
				buf[i++] = tmpbuf[0];
			}
		}
	}

	return i;
}

static void *failover_kick(void *arg) {
	struct failover_dev *dev = (struct failover_dev *)arg;
	char buf[20];
	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	
	if (dev->tty.fd) {
		usleep(20000);
		
		/* open failover */
		write(dev->tty.fd, FAILOVER_DEV_OPEN, 4);
		failover_read(dev, buf);
		failover_read(dev, buf);
		dev->version = atof(buf + 9);
		usleep(50000);

		/* set watchdog kicking interval */
		snprintf(buf, 9, "time%04d", dev->kick_time_interval);
		buf[8] = '\0';
		write(dev->tty.fd, buf, 8);
		failover_read(dev, buf);
	
		/* kick loop */
		failover_event_run(dev, FAILOVER_EVENT_WDOG_OK);

		while (1) {
			write(dev->tty.fd, FAILOVER_DEV_RUNNING, 7);
			failover_event_run(dev, FAILOVER_EVENT_RUNNING);		
			usleep(dev->kick_time_interval * 1000);
		}
	} else {
		failover_event_run(dev, FAILOVER_EVENT_WDOG_FAILED);
	}

	return NULL;
}

void failover_kick_start(struct failover_dev *dev) {
	int ret = -1;
	
	if (!dev)
	{
		return;
	}

	if (dev->running)
		return;

	if (strcasecmp(dev->autorun, ENABLE_AUTORUN)) {
		return;
	}

	dev->running = 1;
	
	/* configure tty device */
	ret = failover_config_tty(dev);

	failover_show(dev);

	/* create pthread */
	if (ret >= 0) {
		pthread_create(&dev->thread, NULL, failover_kick, dev);
	} else {
		failover_kick_stop(dev);
		exit(1);
	}
}

void failover_kick_stop(struct failover_dev *dev) {
	char buf[20];
	
	if (!dev || !dev->running) {
		return;
	}
	dev->running = 0;
	
	
	/* cancel pthread  */
	if (dev->tty.fd >= 0) {
		pthread_cancel(dev->thread);
		pthread_join(dev->thread, NULL);
	}
	
	/* close failover */
	write(dev->tty.fd, FAILOVER_DEV_CLOSE, 5);	
	failover_event_run(dev, FAILOVER_EVENT_STOPPED);

	failover_read(dev, buf);
	usleep(20000);

	/* free tty device */
	failover_free_tty(dev);
}

void failover_show(struct failover_dev *dev) {
	if (dev) {
		if (dev->version) {
			printf(" * DEVICE: %s %s VERSION:%.1f\n",
			dev->dev_name,
			dev->tty.tty_name,
			dev->version);
		} else {
			printf(" * DEVICE: %s %s\n",
				dev->dev_name,
				dev->tty.tty_name);
		}
		
		printf("   INTERVAL:%d EVENT:%s AUTORUN:%s\n",
			dev->kick_time_interval,
			dev->event_cmd,
			dev->autorun);	
	}
}
