/*
 * failover -- The failover cmd tool.
 *
 * Copyright (C) 2009, OpenVox
 *
 * mark.liu <mark.liu@openvox.cn>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include "failover.h"

#define MAX_PATH_LENGTH		256
#define MAX_BUF_LENGHT		256
#define FIFO_RUN_DIR		"/var/run"
#define DEFAULT_FIFO_NAME	"failover.ctl"

/* corresponding getopts_long string */
#define SHORTOPTS	"gnlt:c:d:as"

/* operation modes */
typedef enum {
	MODE_DEBUG = 0,
	MODE_PAUSE,
	MODE_LOGGER,
	MODE_FILE,
	MODE_DEVICE,
	MODE_START,
	MODE_STOP,
	MODE_TIME,
	MODE_HELP,

	MODE_NONE,
} failover_mode_e;

/* mode descriptions */
const static char *longops_descr[MODE_NONE+1][2] = {
	{"Generate debugging output", NULL},
	{"Disable autorun", NULL},
	{"Enable logger", NULL},
	{"Change watchdog timer. Default 4000ms.", "<time>\n\t"},
	{"Set the fifo file", "<fifo_name>\n\t"},
	{"Set the USB device. Default /dev/ttyUSB0.", "<ttyUSB_device_name>\n\t"},
	{"Activate/start failover device. Turn on all leds.", NULL},
	{"Deactivate/shutdown failover device. Turn off all leds.", NULL},
	{"Generate this help statement", NULL},
	{NULL, NULL}
};

/* corresponding getopts_long struct */
const static struct option longopts[MODE_NONE + 1] = {
	{"debug",		0,	NULL,	'g'},
	{"pause",		0,	NULL,	'n'},
	{"logger",		0,	NULL,	'l'},
	{"time",		0,	NULL,	't'},
	{"file",		1,	NULL,	'c'},
	{"device",		1,	NULL,	'd'},
	{"start",		1,	NULL,	'a'},
	{"stop",		0,	NULL,	's'},
	{"help",		0,	NULL,	'h'},

	{ NULL,		0,	NULL,	0},
};

static char fifo_path_name[MAX_PATH_LENGTH];

int debug = 0;

failover_dev_t failover_device;

static void show_usage(FILE *stream, int exit_code) {
	int i;

	fprintf(stream, "Usage: failover [options]\n");
	fprintf(stream, "\tfailover [-gnl] [-c <fifo_file>] [-d <ttyUSB_device_name>]\n");
	fprintf(stream, "\tfailover [-c <file>] -a\n");
	fprintf(stream, "\tfailover [-c <file>] -s\n");
	fprintf(stream, "\tfailover -h\n");

	fprintf(stream, "Examples:\n");

	fprintf(stream, "\tfailover -t 100 -l -d /dev/ttyUSB0\n");
	fprintf(stream, "\t\tfailover -s\n");
	fprintf(stream, "\t\tfailover -a\n");
	fprintf(stream, "\tfailover -c failover1 -d /dev/ttyUSB0\n");
	fprintf(stream, "\tfailover -c failover2 -d /dev/ttyUSB1\n");
	fprintf(stream, "\t\tfailover -c failover1 -s\n");
	fprintf(stream, "\t\tfailover -c failover2 -s\n");
	fprintf(stream, "\t\tfailover -c failover1 -a\n");
	fprintf(stream, "\t\tfailover -c failover2 -a\n");

	fprintf(stream, "Options:\n");
	for (i = 0; longopts[i].name; i ++) {
		fprintf(stream, " -%c --%s %s\t%s\n", longopts[i].val, longopts[i].name, longops_descr[i][1] ? longops_descr[i][1] : "", longops_descr[i][0]);
	}

	exit(exit_code);
}

static int send_cmd(char *cmd) {
	int fd = open(fifo_path_name,O_WRONLY|O_NONBLOCK);
	if(fd < 0) {
		fprintf(stderr, "failover: unable to open %s\n", fifo_path_name);
		pthread_exit(NULL);
	}
	if (debug) {
		printf("open %s success, cmd = %s, len = %d\n", fifo_path_name, cmd, strlen(cmd));
	}

	write(fd, cmd, strlen(cmd));
	close(fd);

	return 0;
}

static void *listen() {
	struct pollfd fds[1];
	int s;
	char buf[MAX_BUF_LENGHT];

	/* open fifo */
	int fd=open(fifo_path_name,O_RDONLY|O_NONBLOCK);
	if(fd < 0) {
		perror("failover: unable to open fifo\n");
		pthread_exit(NULL);
	}
	if (debug) {
		printf("open %s success\n", fifo_path_name);
	}

	memset(buf,0,MAX_BUF_LENGHT);

	while(1) {
		ssize_t bytes;

		memset(buf,0,MAX_BUF_LENGHT);
		fds[0].fd = fd;
		fds[0].events = POLLIN;
		fds[0].revents = 0;

		s = poll(fds, 1, -1);

		if (!(fds[0].revents & POLLIN)) {
			close(fd); /* close fifo */
			fd=open(fifo_path_name,O_RDONLY|O_NONBLOCK);
			continue;
		} 
		
		/* read from fifo */
		bytes=read(fd, buf, MAX_BUF_LENGHT);
		buf[(int)bytes] = 0;
		
		if(bytes > 0) {
			if (debug) {
				printf("fifo receive = %s\n", buf);
			}
			if (!strcmp(buf, FAILOVER_DEV_CLOSE)) {
				failover_kick_stop(&failover_device);
			} 
			else if (!strcmp(buf, FAILOVER_DEV_OPEN)) {
				failover_kick_start(&failover_device);
			}
		}
	}

	close(fd);
	pthread_exit(NULL);
	return NULL;
}

int main(int argc, char *argv[]) {
	char *b_opt_arg;
	char *kick_time_interval;
	int res;
	int interval;
	char *fifoname=NULL;

	failover_init(&failover_device);

	/* scan command line args, croak on invalid args */
	int option;
	while ((option = getopt_long(argc, argv, SHORTOPTS, longopts, NULL)) != EOF) {
		switch(option) {
			case -1: /* Done with options */
				break;

			case '?':
				show_usage(stderr, 1);
				break;

			case 'h':
				show_usage(stdout, 0);

			case 'g': 
				printf("Generate debugging output.\n");
				debug = 1;
				break; 
			case 'n': 
				printf("Disable autorun.\n"); 
				strncpy(failover_device.autorun, DISABLE_AUTORUN, MAX_PARAM_NAME - 1);
				if (debug) {
					printf("autorun = %s\n", failover_device.autorun);
				}
				break; 
			case 'l': 
				if (debug) {
					printf("enable logger\n"); 
				}
				strncpy(failover_device.event_cmd, DEFAULT_EVENT_CMD, MAX_CMD_NAME - 1);
				if (debug) {
					printf("event_cmd = %s\n", failover_device.event_cmd);
				}
				break; 
			case 'c': 
				fifoname = optarg;
				if(failover_strlen_zero(fifoname)) {
					fifoname = DEFAULT_FIFO_NAME; 
				}
				snprintf((char *)fifo_path_name, sizeof(fifo_path_name), "%s/%s", FIFO_RUN_DIR, fifoname);
				break; 
			case 'd': 
				if (NULL == fifoname) {
					fifoname = DEFAULT_FIFO_NAME;
				}
				snprintf((char *)fifo_path_name, sizeof(fifo_path_name), "%s/%s", FIFO_RUN_DIR, fifoname);
				res = mkfifo(fifo_path_name, 0700);
				if (res) {
					if(errno != EEXIST){
						printf("Error creating fifo!\n");
						return -1;
					}
				}

				b_opt_arg = optarg; 

				/* set failover tty_name */
				strncpy(failover_device.tty.tty_name, b_opt_arg, MAX_DEVICE_NAME - 1);
				if (debug) {
					printf("tty device is %s\n", failover_device.tty.tty_name); 
				}

				if (b_opt_arg) {
					failover_kick_start(&failover_device);
					listen();
				}
				break; 
			case 'a': 
				printf("failover start\n"); 
				if(failover_strlen_zero(fifo_path_name)) {
					snprintf((char *)fifo_path_name, sizeof(fifo_path_name), "%s/%s", FIFO_RUN_DIR, DEFAULT_FIFO_NAME);
				}
				send_cmd(FAILOVER_DEV_OPEN);
				break; 
			case 's': 
				printf("failover stop\n"); 
				if(failover_strlen_zero(fifo_path_name)) {
					snprintf((char *)fifo_path_name, sizeof(fifo_path_name), "%s/%s", FIFO_RUN_DIR, DEFAULT_FIFO_NAME);
				}
				send_cmd(FAILOVER_DEV_CLOSE);
				break;
			case 't': 
				kick_time_interval = optarg; 
				interval = atoi(kick_time_interval);
				if((interval < 100) || (interval > 9000)) {
					interval = 4000;
				}

				/* set watchdog kicking interval */
				failover_device.kick_time_interval = interval;
				break;
			case '0':
				break;
			default: /* Something else: unexpected. */
				break;
		}
	}

	return 0;
}
